import 'package:flutter/material.dart';
import 'package:path_stack/path_stack.dart';
import 'nav_stack.dart';
import 'no_animation_delegate.dart';

class NavStackParser extends RouteInformationParser<String> {
  @override
  Future<String> parseRouteInformation(RouteInformation routeInformation) async => routeInformation.location ?? "";

  @override
  RouteInformation? restoreRouteInformation(String path) => RouteInformation(location: path);
}

GlobalKey<NavigatorState> _navKey = GlobalKey();

class NavStackDelegate extends RouterDelegate<String> with ChangeNotifier, PopNavigatorRouterDelegateMixin<String> {
  late final NavStackController controller;
  final String? initialPath;
  final void Function(String value)? onRouteChanging;
  final PathStack Function(BuildContext context, NavStackController controller) onGenerateStack;

  NavStackDelegate({required this.onGenerateStack, this.initialPath, this.onRouteChanging}) {
    // Create a controller that we can make available to the entire app
    controller = NavStackController(
      stackBuilder: onGenerateStack,
      initialPath: initialPath,
      onRouteChanging: onRouteChanging,
    );
    controller.addListener(notifyListeners);
  }

  @override
  GlobalKey<NavigatorState> navigatorKey = _navKey;

  @override
  String? get currentConfiguration => controller.path;

  late String _prevPath = controller.path;

  /// Returns a Navigator with a single page.
  /// Wraps the page in a PathStackPathProvider so the top-most PathStack can grab the current route path.
  /// Calls NavigatorState.widget.stackBuilder to build the page contents.
  @override
  Widget build(BuildContext _) {
    _recordHistoryEntry(controller.path);
    return Material(
      child: Navigator(
        key: navigatorKey,
        transitionDelegate: NoAnimationTransitionDelegate(),
        pages: [
          MaterialPage(
            // Create a context that can access the Navigator above, so the stackBuilder() delegate can show dialogs, overlays etc
            child: Builder(builder: (context) {
              // Allow StackController to be inherited
              return InheritedNavStackController(
                controller: controller,
                // Provide the path to all child PathStack widgets,
                // this saves us the boilerplate of setting .path manually on each instance.
                child: PathStackPathProvider(
                  path: controller.path,
                  // Build the actual stack
                  child: controller.stackBuilder(context, controller),
                ),
              );
            }),
          ),
        ],
        onPopPage: (_, __) => false,
      ),
    );
  }

  // TODO: Should have some support for going up instead of back here. maybe `NavStack.onWillPop(controller)` delegate,
  //  or should it be on the RouteBuilder instead?
  @override
  Future<bool> popRoute() async => controller.goBack();

  @override
  Future<void> setInitialRoutePath(String initialPath) {
    if (initialPath == "/") initialPath = controller.initialPath ?? initialPath;
    return super.setInitialRoutePath(initialPath);
  }

  @override
  Future<void> setNewRoutePath(String path) async => controller.path = path;

  void _recordHistoryEntry(String path) {
    if (path != _prevPath) {
      controller.history.add(path);
    }
    _prevPath = path;
  }
}

class NavStack {
  static NavStackController of(BuildContext context) =>
      context.dependOnInheritedWidgetOfExactType<InheritedNavStackController>()!.controller;
}
